/* File BRMFNS.C  Miscellaneous Broom Balancing functions  */

#ifndef BRMFNS_C
#define BRMFNS_C

#include "brm.h"

state_rec x_to_state_rec (float *x,float x_min,float x_max);
void init_state_rec_factors (void);
void state_rec_to_x (state_rec the_state,float x_min,float x_max,float *x);

/* Implementation  */

state_rec x_to_state_rec (float *x,float x_min,float x_max)
   {
   /*  Assumes x has subscript range 1..4 */
   state_rec the_state;
   float yy[4],*y;
   float xrange_inv;
   int i;
   y = yy - 1;
   xrange_inv = 1.0/(x_max - x_min);
   /* Map x to a variable in the range (0,1) */
   for (i = 1;i <= 4; i++)
       y[i] = (x[i] - x_min)*xrange_inv;
   /* Map to state_rec */
   the_state.x_pos = (x_pos_max - x_pos_min)*y[X_POS_INDEX] + x_pos_min;
   the_state.x_dot = (x_dot_max - x_dot_min)*y[X_DOT_INDEX] + x_dot_min;
   the_state.theta = (theta_max - theta_min)*y[THETA_INDEX] + theta_min;
   the_state.theta_dot = (theta_dot_max - theta_dot_min)*y[THETA_DOT_INDEX]
       + theta_dot_min;
   return the_state;
   }

void init_state_rec_factors (void)
   {
   x_pos_factor = 1.0/(x_pos_max - x_pos_min);
   x_dot_factor = 1.0/(x_dot_max - x_dot_min);
   theta_factor = 1.0/(theta_max - theta_min);
   theta_dot_factor = 1.0/(theta_dot_max - theta_dot_min);
   return;
   }

void state_rec_to_x (state_rec the_state,float x_min,float x_max,float *x)
   {
   /* Maps state_rec to a vector x[1..4] with range x_min to x_max */
   float xrange;
   xrange = x_max - x_min;
   x[X_POS_INDEX] = xrange *(the_state.x_pos - x_pos_min) * x_pos_factor
       + x_min;
   x[X_DOT_INDEX] = xrange *(the_state.x_dot - x_dot_min) * x_dot_factor
       + x_min;
   x[THETA_INDEX] = xrange *(the_state.theta - theta_min) * theta_factor
       + x_min;
   x[THETA_DOT_INDEX] = xrange *(the_state.theta_dot - theta_dot_min) *
       theta_dot_factor + x_min;
   return;
   }

#endif
