/* File DYNFFRUN.C Run a feedforward neural network as a */
/*      nonlinear dynamical system */

#ifndef DYNFFRUN_C
#define DYNFFRUN_C

#include <conio.h>

#include "trn.h"
#include "dyn.h"
#include "ff.h"

void run_dynamic_network (ff_setup_rec setup,dyn_setup_rec dyn_setup);

/* Implementation */

network_rec dynamic_net;  /* used in this file only */

void neural_network_dyn_sys (float x_in,float y_in,float *x_out,float *y_out)
     {
     /*  Normalization routines and feedforward assume array indices */
     /*  starting at 1 (not 0), so we 'cheat' here, and allow one extra  */
     /*  array position.  */
     float net_input[3],net_output[3];
     net_input [1] = x_in;
     net_input [2] = y_in;
     normalize_data (net_input,dynamic_net.x_inp,2,g_tr_rec.max_in,
	g_tr_rec.min_in,g_ff_setup.norm_in_range,g_ff_setup.norm_in_min);
     feedforward (dynamic_net,g_system_weights,g_ff_setup);
     denormalize_output (dynamic_net.y_out,net_output,2,
	g_ff_setup.norm_out_range,g_ff_setup.norm_out_min,g_tr_rec);
     *x_out = net_output [1];
     *y_out = net_output [2];
     return;
     }  /* end proc */

void run_dynamic_network (ff_setup_rec setup,dyn_setup_rec dyn_setup) {
     clrscr();
     textcolor (YELLOW);
     gotoxy (1,3);
     cputs ("Neural Net Dynamical System Program");
     gotoxy (1,5);
     if (setup.files.in_wts_file_name[0] == '\0') {
	cprintf ("\r\nNetwork Weights file name is blank.\r\n");
	cputs ("Press key to cancel: ");
	getch();
	return;
	}
     if (setup.files.training_file_name[0] == '\0') {
	cprintf ("\r\nTraining file name is blank.\r\n");
	cputs ("Press key to cancel: ");
	getch();
	return;
	}
     cprintf("Input weights file name: %s",setup.files.in_wts_file_name,'.');
     gotoxy (1,7);
     cprintf ("Training file name: %s.",setup.files.training_file_name);
     gotoxy (1,9);
     cputs ("Press any key to continue (esc to cancel): ");
     if (getch() == ESC_KEY) return;
     clrscr();
     if (!process_training_file (setup.files.training_file_name,
	setup.no_of_inputs,setup.no_of_outputs,&g_tr_rec))
	goto exit_proc;
     if (!read_weights (&g_system_weights,setup)) goto exit_proc;
     allocate_network (setup,&dynamic_net);
     graph_dynamical_system (&(dyn_setup.gr_rec),&neural_network_dyn_sys);
     free_network (&dynamic_net);
exit_proc:
     free_training_pointers(&g_tr_rec);
     free_weights (&g_system_weights);
     return;
     }  /* end proc  */

#endif