/* File FFWTS.C  Allocate, read in, write out feedforward network weights */

#ifndef FFWTS_C
#define FFWTS_C

#include <stdlib.h>
#include <conio.h>
#include <stdio.h>
#include <alloc.h>

#include "ut.h"
#include "ff.h"

/* Note:  All weight indices start with 1  */

void move_2d_double (double **source,double **dest,int no_of_rows,
   int no_of_cols);
weights_rec allocate_network_weights (ff_setup_rec setup);
void free_weights (weights_rec *weights);
short read_weights (weights_rec *weights,ff_setup_rec setup);
short write_weights (weights_rec weights,path_str file_name,
	 ff_setup_rec setup);
short print_weights (weights_rec weights,path_str file_name,
	 ff_setup_rec setup);
weights_rec initialize_weights (ff_setup_rec setup,short random_init);

/* Implementation  */

#define FIELD_LEN 80

short check_wts_parameters (ff_setup_rec setup,FILE *infile,
   int line) {
   /* Returns 1 if check is OK, 0 otherwise */
   int n_in,n_h1,n_out,n_sig_out;
   float n_in_min,n_out_min;
   char  char_str[FIELD_LEN];
   gotoxy (1,line);
   fgets (char_str,FIELD_LEN,infile);
   if (sscanf (char_str,"%d %d %d %f %f %d",&n_in,&n_h1,&n_out,&n_in_min,
       &n_out_min,&n_sig_out) != 6) {
       cprintf ("Problem reading in parameters in weights file.");
       getch();
       return 0;
       }
   cprintf ("No. of input nodes: %d",n_in);
   gotoxy (1,line + 2);
   cprintf ("No. of hidden nodes: %d",n_h1);
   gotoxy (1,line + 4);
   cprintf ("No. of output nodes: %d",n_out);
   gotoxy (1,line + 6);
   cprintf ("Normalized Input Min: %7.4f",n_in_min);
   gotoxy (1,line + 8);
   cprintf ("Normalized Output Min: %7.4f",n_out_min);
   gotoxy (1,line + 10);
   if (n_sig_out)
      cprintf ("Sigmoidal Output Layer\r\n");
   else
      cprintf ("Linear Output Layer\r\n");
   if ((n_in != setup.no_of_inputs) ||
       (n_h1 != setup.no_of_h1_nodes) ||
       (n_out != setup.no_of_outputs) ||
       (n_in_min != setup.norm_in_min) ||
       (n_out_min != setup.norm_out_min) ||
       (n_sig_out != setup.sigmoid_out)) {
	  gotoxy (1,line + 12);
	  cprintf ("These numbers do not agree with setup.");
	  cprintf ("\r\nPress any key to cancel: ");
	  getch();
	  return 0;
	  }
    return 1;  /* Parameters are OK */
    }  /* end proc */

double **allocate_2d_double (int rows_upper_lim,int cols_upper_lim) {
   /* Allocate a 2D double weight matrix with subscript range */
   /* 1..rows_upper_lim,1..cols_upper_lim */
   #define rows_lower_lim 1
   #define cols_lower_lim 1
   int i,no_of_rows = rows_upper_lim - rows_lower_lim + 1,
       no_of_cols = cols_upper_lim - cols_lower_lim + 1;
   double **a;
   /* Allocate pointers to rows */
   a = (double **) malloc((size_t)((no_of_rows+1)*sizeof(double*)));
   if (!a) {
      clrscr();
      gotoxy (1,5);
      cprintf ("Memory allocation 1 failed in allocate_wts");
      getch();
      exit(1);
      }
   a -= rows_lower_lim;
   /* Allocate rows and set pointers to them */
   a[rows_lower_lim] = (double *) malloc
     ((size_t)((no_of_rows+1)*(no_of_cols+1))*sizeof(double));
   if (!a[rows_lower_lim]) {
      clrscr();
      gotoxy (1,5);
      cprintf ("Memory allocation 2 failed in allocate_wts");
      getch();
      exit(1);
      }
   a[rows_lower_lim] -= cols_lower_lim;
   for (i=rows_lower_lim + 1;i <= rows_upper_lim; i++)
       a[i] = a[i-1] + no_of_cols;
   /* Return pointer to array of pointers to rows */
   return a;
   }

void free_2d_double (double **a) {
   if (a[1]) free ((char *)(a[1]));
   if (a) free ((char *)(a));
   }

void move_2d_double (double **source,double **dest,int no_of_rows,
   int no_of_cols) {
   /*  Assumes starting index of 1 in each dimension */
   int i,j;
   for (i=1;i<=no_of_rows;i++)
       for (j=1;j<=no_of_cols;j++)
	   dest [i][j] = source[i][j];
   return;
   }

weights_rec allocate_network_weights (ff_setup_rec setup) {
   weights_rec wts;
   wts.wi = allocate_2d_double(setup.no_of_inputs + 1,
		 setup.no_of_h1_nodes);
   wts.wo = allocate_2d_double(setup.no_of_h1_nodes,setup.no_of_outputs);
   wts.allocated = TRUE;
   return wts;
   }

void free_weights (weights_rec *weights) {
   if (weights->allocated) {
      free_2d_double (weights->wi);
      free_2d_double (weights->wo);
      }
   weights->allocated = FALSE;
   return;
   }

short read_weights (weights_rec *weights,ff_setup_rec setup) {
  /* Returns 1 if file read is OK, 0 otherwise */
   int i,j;
   FILE *infile;
   const display_line = 7;
   clrscr();
   if (!weights->allocated)
      *weights = allocate_network_weights (setup);
   gotoxy (1,5);
   cprintf ("Reading weights file...");
   if (!open_input_text_file (&infile,setup.files.in_wts_file_name)) return 0;
   if (!check_wts_parameters (setup,infile,display_line)) return 0;
   for (j=1;j<=setup.no_of_h1_nodes;j++) {
       for (i=1;i<=setup.no_of_inputs;i++)
	   fscanf (infile,"%lf",&(weights->wi[i][j]));
       /* Threshold node */
       fscanf (infile,"%lf\n",&(weights->wi[setup.no_of_inputs + 1][j]));
       } /* j  */

   fscanf (infile,"\n");  /* Skip a line to separate input and output */

   for (j = 1;j<=setup.no_of_outputs;j++) {
       for (i=1;i<=setup.no_of_h1_nodes;i++)
	   fscanf (infile,"%lf",&(weights->wo[i][j]));
       fscanf (infile,"\n");
       }

   fclose (infile);
   return 1;
   }  /* end proc */

short write_weights (weights_rec weights,path_str file_name,
	 ff_setup_rec setup) {
/* Define precision format for output (15 significant digits) */
#define PREC_FORMAT "%21.14lE "
   int i,j;
   FILE *outfile;
   if (!open_output_text_file (&outfile,file_name)) return 0;
   gotoxy (1,12);
   cprintf ("Writing weights to file...");
   fprintf (outfile,"%3d  %3d  %3d  %7.4f  %7.4f  %d\n",setup.no_of_inputs,
       setup.no_of_h1_nodes,setup.no_of_outputs,setup.norm_in_min,
       setup.norm_out_min,setup.sigmoid_out);
   /*  Write weights with 15 significant digits in exponential form  */
   for (j=1;j<= setup.no_of_h1_nodes;j++) {
       for (i=1;i<=setup.no_of_inputs;i++)
	   fprintf (outfile,PREC_FORMAT,weights.wi[i][j]);
       /* Threshold node: */
       fprintf (outfile,PREC_FORMAT "\n",weights.wi[setup.no_of_inputs + 1][j]);
       }  /* end j  */
   fprintf (outfile,"\n");  /* Skip a line to separate input and output */
   for (j = 1;j<=setup.no_of_outputs;j++) {
       for (i = 1;i<=setup.no_of_h1_nodes;i++)
	   fprintf (outfile,PREC_FORMAT,weights.wo[i][j]);
       fprintf (outfile,"\n");
       }
   fclose (outfile);
   gotoxy (1,12);
   /*  Clear message: */
   cprintf ("                          ");
   return 1;
   }  /* end proc */

short print_weights (weights_rec weights,path_str file_name,
	 ff_setup_rec setup) {

#define PRINT_FORMAT "%9.6f "
/* Number of items printed across page:  */
#define PRINT_WIDTH 5

   int i,j,prt_start,prt_end,prt_items,remaining_items;
   fprintf (stdprn,"\r\n\nWeights File: %s\n",file_name);
   fprintf (stdprn,"%3d  %3d  %3d  %7.4f  %7.4f  %d\n",setup.no_of_inputs,
       setup.no_of_h1_nodes,setup.no_of_outputs,setup.norm_in_min,
       setup.norm_out_min,setup.sigmoid_out);
   fputs ("Input Weights: \n",stdprn);
   for (j=1;j<= setup.no_of_h1_nodes;j++) {
       remaining_items = setup.no_of_inputs;
       prt_start = 1;
       do {
       if (remaining_items < PRINT_WIDTH) prt_items = remaining_items;
       else prt_items = PRINT_WIDTH;
       prt_end = prt_start + prt_items - 1;
       fprintf (stdprn,"%2d, %2d  ",i,j);
       for (i=prt_start;i<=prt_end;i++) {
	   fprintf (stdprn,PRINT_FORMAT,weights.wi[i][j]);
	   remaining_items--;
	   } /*  i  */
       fputs ("\n",stdprn);
       prt_start = prt_end + 1;
       } while (remaining_items > 0);
       fprintf (stdprn,"%2d, %2d  ",setup.no_of_inputs + 1,j);
       /* Threshold node: */
       fprintf (stdprn,PRINT_FORMAT "\n",weights.wi[setup.no_of_inputs + 1][j]);
       }  /* end j  */
   fputs ("Output Weights: \n",stdprn);
   for (j = 1;j<=setup.no_of_outputs;j++) {
       remaining_items = setup.no_of_h1_nodes;
       prt_start = 1;
       do {
       if (remaining_items < PRINT_WIDTH) prt_items = remaining_items;
       else prt_items = PRINT_WIDTH;
       prt_end = prt_start + prt_items - 1;
       fprintf (stdprn,"%2d, %2d  ",i,j);
       for (i=prt_start;i<=prt_end;i++) {
	   fprintf (stdprn,PRINT_FORMAT,weights.wo[i][j]);
	   remaining_items--;
	   } /*  i  */
       fputs ("\n",stdprn);
       prt_start = prt_end + 1;
       } while (remaining_items > 0);
       } /*  j  */
   return 1;
   }  /* end proc */

weights_rec initialize_weights (ff_setup_rec setup,short random_init) {
   /*  Allocate and randomly initialize weights  */
   int i,j;
   weights_rec weights;
   weights = allocate_network_weights (setup);
   clrscr();
   gotoxy (1,5);
   cprintf ("Initializing weights...");
   gotoxy (1,7);
   cprintf ("Input: ");
   for (i = 1;i<= (setup.no_of_inputs + 1);i++) {
       for (j = 1;j<=setup.no_of_h1_nodes;j++) {
	   if (random_init) weights.wi[i][j] = setup.weight_range *
		(2*frand() - 1.0); /* Random between -1 and 1 */
	   else weights.wi[i][j] = 0.0;
	   } /* j */
       gotoxy (8,7);
       cprintf ("              ");
       gotoxy (8,7);
       cprintf ("%2d of %d done.",i,setup.no_of_inputs);
       } /* i */
   gotoxy (1,9);
   cprintf ("Output: ");
   for (i = 1;i<=setup.no_of_h1_nodes;i++) {
       for (j = 1;j<=setup.no_of_outputs;j++) {
	   if (random_init) weights.wo[i][j] = setup.weight_range *
		(2*frand() - 1.0); /* Random between -1 and 1 */
	   else weights.wo [i][j] = 0.0;
	   } /* j */
       gotoxy (8,9);
       cprintf ("              ");
       gotoxy (8,9);
       cprintf ("%2d  of %d done.",i,setup.no_of_h1_nodes);
       } /* i */
   return weights;
   } /* proc */

#endif