/*  File FLGINIT.C Initialization functions for Fuzzy Logic-Genetic Modeling */

#ifndef FLGINIT_C
#define FLGINIT_C

#include <alloc.h>
#include <string.h>

#include "flga.h"

void init_flga_setup (tflg_setup_rec *fl) {
   fl->no_of_fl_inputs = 5;
   fl->no_of_fl_inp_regions = 3;
   fl->no_of_fl_output_values = 8;
   fl->read_in_fuzzy_system = 0;
   fl->norm_in_range = 5.0;
   fl->norm_in_min = -2.5;
   fl->norm_out_range = 2.0;
   fl->norm_out_min = -1.0;
   strcpy(fl->fuzzy_system_file_name,"");
   strcpy(fl->training_file_name,"");
   strcpy(fl->out_data_file_name,"");
   return;
   }  /*  end proc  */

void init_flga_rules (fuzzy_system_rec *fz) {
   int i0,i1,i2,i3,i4,i = 0,j,n = 1,out_value = fz->no_of_outputs/2;
   for (j = 1;j<=fz->no_of_inputs;j++)
       n *= fz->no_of_inp_regions;
   fz->no_of_rules = n;
   fz->rules = (rule *) malloc ((size_t)(fz->no_of_rules*sizeof(rule)));
   for (i0 = 0;i0 < fz->no_of_inp_regions;i0++)
       for (i1 = 0;i1 < fz->no_of_inp_regions;i1++)
	   for (i2 = 0;i2 < fz->no_of_inp_regions;i2++)
	       for (i3 = 0;i3 < fz->no_of_inp_regions;i3++)
		   for (i4 = 0;i4 < fz->no_of_inp_regions;i4++) {
			for (j=0;j<fz->no_of_inputs;j++)
			    fz->rules[i].inp_index[j] = j;
			    fz->rules[i].inp_fuzzy_set[0] = i0;
			    fz->rules[i].inp_fuzzy_set[1] = i1;
			    fz->rules[i].inp_fuzzy_set[2] = i2;
			    fz->rules[i].inp_fuzzy_set[3] = i3;
			    fz->rules[i].inp_fuzzy_set[4] = i4;
			    fz->rules[i].out_fuzzy_set = out_value;
			    i++;
			    }  /*  end loop */
   return;
   }  /*  end proc */

void init_flga_mem_fns (tflg_setup_rec fl,fuzzy_system_rec *fz) {
   int i,j,k;
   float *a,a_inc;
   int no_of_rgns,no_of_pts = 2 * fl.no_of_fl_inp_regions;
   no_of_rgns = no_of_pts - 1;
   a_inc = fl.norm_in_range / no_of_rgns;
   a = (float *)malloc(no_of_pts*sizeof(float));
   a[0] = fl.norm_in_min;
   for (i = 1;i<no_of_pts;i++)
       a[i] = a[i-1] + a_inc;
   /* The input membership functions */
   for (i = 0;i<fl.no_of_fl_inputs;i++) {
       fz->inp_mem_fns[i][0] = init_trapz (a[1],a[2],0,0,left);
       k = 1;
       for (j = 1;j<(fl.no_of_fl_inp_regions - 1);j++) {
	   fz->inp_mem_fns[i][j] =
	       init_trapz (a[k],a[k+1],a[k+2],a[k+3],regular);
	   k += 2;
	   } /* end j */
       fz->inp_mem_fns[i][fl.no_of_fl_inp_regions-1] =
	   init_trapz (a[k],a[k+1],0,0,right);
       }  /* end i */
   return;
   }  /* end proc */

void init_flga_fuzzy_system (tflg_setup_rec fl,fuzzy_system_rec *fz) {
     float out_inc, out_val;
     int i;
     fz->no_of_inputs = 5;
     fz->no_of_inp_regions = 3;
     fz->no_of_outputs = 8;
     init_flga_mem_fns(fl,fz);
     init_flga_rules(fz);
     out_inc = fl.norm_out_range/(fz->no_of_outputs - 1);
     out_val = fl.norm_out_min;
     for (i = 0;i<fz->no_of_outputs;i++) {
	 fz->output_values [i] = out_val;
	 out_val += out_inc;
	 }  /*  end i */
     fz->allocated = TRUE;
     return;
     }  /*  end proc  */

#endif