/*  File GAPROCS.C  Functions and procedures for genetic algorithm */

#ifndef GAPROCS_C
#define GAPROCS_C

#include <stdlib.h>
#include <conio.h>
#include <alloc.h>
#include <mem.h>
#include <math.h>

#include "uttypes.h"
#include "ut.h"
#include "ga.h"

#define TOO_SMALL 1e-6    /* Used for fitness scaling  */

tga_setup_rec ga_setup_rec;
tga_rec ga_rec;

float power_of_2 (int n);
void initialize_ga_rec (tga_rec *ga);
short initialize_genetic_alg (tga_setup_rec *setup,tga_rec *ga,
   float (*the_obj_funct)(ui_vector v));
void generation (tga_setup_rec *setup,tga_rec *ga,
   float (*the_obj_funct)(ui_vector v));
void statistics (tga_setup_rec setup,ppopulation the_population,tga_rec *ga,
   short *new_opt);
void display (int gen,tga_setup_rec setup,tga_rec ga,short *cancel,
   short *toler_flag);
void move_individual (tga_setup_rec setup,individual src_ind,
   individual *dest_ind);
void free_ga_pointers (tga_setup_rec setup,tga_rec *ga);
void initialize_ga_setup (int pop_size,int bit_len,int vect_len,
  tga_setup_rec *the_rec);

/* Implementation  */

void initialize_ga_rec (tga_rec *ga) {
     ga->allocated = 0;
     ga->populations[0] = NULL;
     ga->populations[1] = NULL;
     ga->opt_individual.chrom = NULL;
     ga->opt_individual.raw_fitness = 0.0;
     ga->opt_individual.scaled_fitness = 0.0;
     ga->opt_individual.parent1 = 0;
     ga->opt_individual.parent2 = 0;
     ga->opt_individual.xsite = 0;;
     ga->opt_vector = NULL;
     ga->g_old = 0;
     ga->g_new = 1;
     ga->opt_fitness = 0.0;
     ga->sum_fitness = 0.0;
     ga->nmutation = 0;
     ga->ncross = 0;
     ga->avg = 0.0;
     ga->max = 0.0;
     ga->min = 0.0;
     return;
     }

float power_of_2 (int n) {
   int i;
   float val;
   val = 1.0;
   if (n > 0) {
      for (i = 1;i <= n;i++)
	  val *= 2.0;
      return val;
      }
   if (n < 0) {
      for (i = 1;i <= n;i++)
	  val *= 0.5;
      }
   return val;
   }  /*  end fn */

short flip (float prob) {
   /* prob is a probability (i.e., between 0 and 1) */
   /* frand() is defined in UT.H */
   if (frand() < prob) return 1;
   return 0;
   }  /* end fn */

int rnd(int low_lim,int up_lim) {
   /* return random between low_lim and up_lim  */
   int range,result;
   range = up_lim - low_lim + 1;
   result = random (range);  /* Result is between 0 and range - 1 */
   return  (result + low_lim);
   }  /* end fn */

int select (int popsize, float sumfitness,ppopulation the_population) {
   float rand,sum = 0.0;
   int j = 0;
   rand = frand() * sumfitness;
   do {
      sum += (*the_population)[j]->scaled_fitness;
      j++;
      } while ((sum < rand) && (j < popsize));
   return (j - 1);
   }  /* end fn */

short mutation (short val,float pmutation,long int *nmutation) {
    short  mutate;
    mutate = flip (pmutation);
    if (mutate) {
       (*nmutation)++;
       return !val;
       }
    else
       return val;
    }  /* end fn */

void crossover (chromosome parent1,chromosome parent2,
    chromosome child1,chromosome child2,
   int lchrom,long int *ncross,long int *nmutation,int *jcross,
   float pcross,float pmutation) {
   /* Cross two parent strings, place in two child strings */
   int j;
   if (flip (pcross)) {
      *jcross = rnd (0,lchrom - 2);
      (*ncross)++;
      }
   else
      *jcross = lchrom - 1;
   for (j = 0;j <= *jcross;j++) {
      child1 [j] = mutation (parent1 [j],pmutation,nmutation);
      child2 [j] = mutation (parent2 [j],pmutation,nmutation);
      }  /* end j */
   if (*jcross != (lchrom - 1))
      for (j = *jcross + 1;j <= lchrom - 1;j++) {
	  child1 [j] = mutation (parent2 [j],pmutation,nmutation);
	  child2 [j] = mutation (parent1 [j],pmutation,nmutation);
	  }  /* end j */
   return;
   }  /* end proc */

float scale_fitness (float x,float fmin,float fmax) {
   if (fabs (fmax - fmin) > TOO_SMALL)
      return (x - fmin)*0.95/(fmax - fmin) + 0.05;
 /*   else  */
   return 1.0;
   } /* end fn */

void rescale_population_fitness (tga_setup_rec setup,
   tga_rec ga,ppopulation the_population) {
   int j;
   float max_fitness,min_fitness;
   min_fitness = (*the_population)[0]->raw_fitness;
   max_fitness = (*the_population)[0]->raw_fitness;
   for (j = 0;j < setup.population_size;j++) {
       if ((*the_population)[j]->raw_fitness < min_fitness)
		  min_fitness = (*the_population)[j]->raw_fitness;
       if ((*the_population)[j]->raw_fitness > max_fitness)
		  max_fitness = (*the_population)[j]->raw_fitness;
       }
  if (ga.opt_individual.raw_fitness > max_fitness)
      max_fitness = ga.opt_individual.raw_fitness;
  if (max_fitness - min_fitness < TOO_SMALL)
      min_fitness = 0.9 * max_fitness;
  for (j = 0;j < setup.population_size;j++)
       (*the_population)[j]->scaled_fitness =
	 scale_fitness ((*the_population)[j]->raw_fitness,
	 min_fitness,max_fitness);
  return;
  }  /* end proc */

void pick_fittest (tga_setup_rec setup,
   ppopulation the_old,ppopulation the_new) {
   int i,j,min_index = 0;
   float min_fitness;
   /* Replace min fitness members of new population with fitter members */
   /* from old population.  */
   min_fitness = (*the_new)[0]->raw_fitness;
   for (j = 1;j<setup.population_size;j++)
       if ((*the_new)[j]->raw_fitness < min_fitness) {
	  min_fitness = (*the_new)[j]->raw_fitness;
	  min_index = j;
	  }
   for (i = 0;i<setup.population_size;i++)
       if ((*the_old)[i]->raw_fitness > min_fitness) {
	  /* Replace min individual with new individual */
	  move_individual (setup,*(*the_old)[i],
	       (*the_new)[min_index]);
	  /* Determine min of new population */
	  min_fitness = (*the_new)[0]->raw_fitness;
	  min_index = 0;
	  for (j = 1;j<setup.population_size;j++)
	      if ((*the_new)[j]->raw_fitness < min_fitness) {
		  min_fitness = (*the_new)[j]->raw_fitness;
		  min_index = j;
		  }  /* end if, j */
	  }  /* if,i */
    return;
    } /* end proc */

void decode (tga_setup_rec setup, chromosome chrom,ui_vector v) {
   /* Decode from chromosome to vector */
   int i,k,count = 0;
   unsigned int accum,apower_of_2;
   int start_pos = 0,end_pos = setup.bit_len;
   for (i = 0;i<setup.vect_len;i++) {
	accum = 0;
	apower_of_2 = 1;
	for (k = start_pos;k < end_pos;k++) {
	    if (chrom [count])
	       accum += apower_of_2;
	       count += 1;
	       apower_of_2 *= 2;
	       } /* end k */
	v[i] = accum;
	start_pos += setup.bit_len;
	end_pos += setup.bit_len;
	} /* end i */
   return;
   }  /* end proc */

void generation (tga_setup_rec *setup,tga_rec *ga,
   float (*the_obj_funct)(ui_vector v)) {
   /* Note: Population size must be even  */
   int j = 0,k,mate1,mate2,jcross;
   ui_vector v;
   v = allocate_ui_vector (0,setup->vect_len - 1);
   do {
      mate1 = select (setup->population_size,
	      ga->sum_fitness,ga->populations[ga->g_old]);
      mate2 = select (setup->population_size,
	      ga->sum_fitness,ga->populations[ga->g_old]);
      crossover ((*(ga->populations)[ga->g_old])[mate1]->chrom,
		 (*(ga->populations)[ga->g_old])[mate2]->chrom,
	    (*(ga->populations)[ga->g_new])[j]->chrom,
	    (*(ga->populations)[ga->g_new])[j+1]->chrom,
	    setup->chrom_len,&(ga->ncross),&(ga->nmutation),&jcross,
	    setup->crossover_prob,
	    setup->mutation_prob);
      for (k = j;k <= j + 1;k++) {
	  decode (*setup,(*(ga->populations)[ga->g_new])[k]->chrom,v);
	  (*(ga->populations)[ga->g_new])[k]->raw_fitness =
		  the_obj_funct (v);
	  (*(ga->populations)[ga->g_new])[k]->parent1 = mate1;
	  (*(ga->populations)[ga->g_new])[k]->parent2 = mate2;
	  (*(ga->populations)[ga->g_new])[k]->xsite = jcross;
	  }
      j += 2;
    } while (j < setup->population_size);
    free_ui_vector (v,0);
    if (setup->pick_fittest)
       pick_fittest (*setup,(ga->populations)[ga->g_old],
		    (ga->populations)[ga->g_new]);
    rescale_population_fitness (*setup,*ga,ga->populations[ga->g_new]);
    return;
    } /* end proc */

void statistics (tga_setup_rec setup,ppopulation the_population,tga_rec *ga,
   short *new_opt) {
   int j;
   *new_opt = FALSE;
   ga->sum_fitness = (*the_population)[0]->scaled_fitness;
   ga->min = (*the_population)[0]->raw_fitness;
   ga->max = (*the_population)[0]->raw_fitness;
   if (ga->max > ga->opt_individual.raw_fitness) {
      move_individual (setup,*(*the_population)[0],&(ga->opt_individual));
      *new_opt = TRUE;
      }
   for (j = 1;j<setup.population_size;j++) {
       ga->sum_fitness += (*the_population)[j]->scaled_fitness;
       if ((*the_population)[j]->raw_fitness > ga->max)
	     ga->max = (*the_population)[j]->raw_fitness;
       if ((*the_population)[j]->raw_fitness < ga->min)
	     ga->min = (*the_population)[j]->raw_fitness;
       if (ga->max > (ga->opt_individual).raw_fitness)
	  move_individual (setup,*(*the_population)[j],&(ga->opt_individual));
       } /* end j */
   if ((ga->opt_individual).raw_fitness > ga->opt_fitness) {
      decode (setup,ga->opt_individual.chrom,ga->opt_vector);
      ga->opt_fitness = ga->opt_individual.raw_fitness;
      *new_opt = TRUE;
      }
   ga->avg = ga->sum_fitness/setup.population_size;
   return;
   }  /* end proc */

void display_vector (ui_vector v,int vlen) {
   int i;
   int display_len = vlen;
   const max_display_len = 8;
   if (display_len > max_display_len)
      display_len = max_display_len;
   cprintf ("\r\n");
   for (i = 0;i<display_len;i++)
       cprintf("%6u ",v[i]);
   cprintf ("\r\n");
   return;
   }  /* end proc */

void display (int gen,tga_setup_rec setup,tga_rec ga,short *cancel,
   short *toler_flag) {
   int j;
   *toler_flag = FALSE;
   clrscr();
   cprintf ("Gen: %3d\r\n",gen);
   for (j = 0;j<setup.population_size;j++) {
       cprintf ("\r\n%3d: (%3d,%3d)  %3d  ",
		   j,(*ga.populations[ga.g_new])[j]->parent1,
		   (*ga.populations[ga.g_new])[j]->parent2,
		   (*ga.populations[ga.g_new])[j]->xsite);
       cprintf ("Fitness: %6.4f,  Scaled: %6.4f",
		(*ga.populations[ga.g_new])[j]->raw_fitness,
		(*ga.populations[ga.g_new])[j]->scaled_fitness);
       }  /* j */
   cprintf ("\r\nGeneration %3d\r\n",gen);
   cprintf ("Max: %6.4f, Min: %6.4f, Scaled Avg: %6.4f\r\n",
      ga.max,ga.min,ga.avg);
   cprintf ("Sum: %6.4f, Nmutation: %ld, Ncross: %ld\r\n\n",
      ga.sum_fitness,ga.nmutation,ga.ncross);
   if (fabs (MAX_GA_OBJ_FN_VALUE - ga.max) < setup.tolerance) {
      *toler_flag = TRUE;
      cprintf ("\r\nThe following chromosomes are within fitness tolerance "
	    "(%10.8f): \r\n",setup.tolerance);
      for (j = 0;j< setup.population_size;j++)
	  if (fabs (MAX_GA_OBJ_FN_VALUE -
	     (*ga.populations[ga.g_new])[j]->raw_fitness) <
		setup.tolerance)
	     cprintf ("%2d: Fitness: %6.4f\r\n",j,
		(*ga.populations[ga.g_new])[j]->raw_fitness);
      }
   cprintf("\r\nOptimum Individual Fitness: %7.4f",
	  ga.opt_individual.raw_fitness);
   cprintf("\r\nOptimum Vector Fitness: %7.4f",ga.opt_fitness);
   display_vector (ga.opt_vector,setup.vect_len);
   cprintf ("\r\nPress esc to cancel: ");
   if (kbhit() && (getch () == ESC_KEY)) *cancel = TRUE;
   else *cancel = FALSE;
   return;
   }  /* end proc */

void display_stats (int gen, tga_setup_rec setup,tga_rec ga) {
   clrscr();
   cprintf ("\r\n");
   cprintf ("Generation %2d\r\n",gen);
   cprintf ("Max: %6.4f, Min: %6.4f, Avg: %6.4f\r\n",
      ga.max,ga.min,ga.avg);
   cprintf ("Sum: %6.4f, Nmutation: %d, Ncross: %d\r\n\n",
      ga.sum_fitness,ga.nmutation,ga.ncross);
   if (fabs (MAX_GA_OBJ_FN_VALUE - ga.opt_individual.raw_fitness)
	    < setup.tolerance) {
      cprintf ("\r\nOptimum individual is within fitness tolerance  "
	    "(%10.8f): \r\n:",setup.tolerance);
      cprintf ("Opt Individual Fitness: %6.4f\r\n",
	 ga.opt_individual.raw_fitness);
      }  /* end if */
    return;
    }   /* end proc */

void move_individual (tga_setup_rec setup,individual src_ind,
   individual *dest_ind) {
   /* Move content of src_ind to dest_ind */
   /*  Need to move contents of chrom, not the pointer value chrom ! */
   memmove(dest_ind->chrom,src_ind.chrom,
       (sizeof(src_ind.chrom[0]))*setup.chrom_len);
   dest_ind->raw_fitness = src_ind.raw_fitness;
   dest_ind->scaled_fitness = src_ind.scaled_fitness;
   dest_ind->parent1 = src_ind.parent1;
   dest_ind->parent2 = src_ind.parent2;
   dest_ind->xsite = src_ind.xsite;
   return;
   }  /*  end proc  */

short allocate_population (tga_setup_rec setup,
   ppopulation the_population) {
   int j;
   *the_population =
      (population)malloc(setup.population_size*sizeof(pindividual));
   if (!(*the_population)) return 0;
   for (j = 0;j<setup.population_size;j++) {
       (*the_population)[j] = (pindividual) malloc(sizeof(individual));
       if (!(*the_population)[j]) return 0;
       (*the_population)[j]->chrom =
		(short *)malloc(setup.chrom_len*sizeof(short));
       if (!((*the_population)[j]->chrom)) return 0;
       }
   return 1;
   } /*  end proc  */

void free_population (tga_setup_rec setup,
   ppopulation the_population) {
   int j;
   if (the_population)
   for (j = 0;j<setup.population_size;j++) {
       if ((*the_population)[j]->chrom) free ((*the_population)[j]->chrom);
       if ((*the_population)[j]) free ((*the_population)[j]);
       }
   return;
   } /*  end proc  */

void initialize_population (tga_setup_rec setup,tga_rec *ga,
   ppopulation the_population, pindividual the_opt,
   float (*the_obj_funct)(ui_vector v)) {
   int i,j,k;
   float max_fitness;
   int opt_index;
   ui_vector v;
   v = allocate_ui_vector (0,setup.vect_len - 1);
   /*  Initialize ga.opt_vector  */
   ga->opt_vector = allocate_ui_vector (0,setup.vect_len - 1);
   for (j = 0;j<setup.population_size;j++) {
       for (k = 0;k < setup.chrom_len;k++)
	   (*the_population)[j]->chrom [k] = flip (0.5);
       decode (setup,(*the_population)[j]->chrom,v);
       (*the_population)[j]->raw_fitness = the_obj_funct (v);
       (*the_population)[j]->scaled_fitness = 0.0;
       (*the_population)[j]->parent1 = 0;
       (*the_population)[j]->parent2 = 0;
       (*the_population)[j]->xsite = 0;
       cprintf ("\r\nIndividual %2d initialized.",j);
       }  /* end j */
   max_fitness = (*the_population)[0]->raw_fitness;
   opt_index = 0;
   for (j = 1;j < setup.population_size;j++)
       if (max_fitness < (*the_population)[j]->raw_fitness) {
	  max_fitness = (*the_population)[j]->raw_fitness;
	  opt_index = j;
	  }
   move_individual(setup,*((*the_population)[opt_index]),the_opt);
   decode (setup,the_opt->chrom,ga->opt_vector);
   ga->opt_fitness = max_fitness;
   rescale_population_fitness (setup,*ga,the_population);
   free_ui_vector (v,0);
   return;
   }  /* end proc */

void initial_display (tga_setup_rec setup,tga_rec ga,short *cancel){
   clrscr();
   textcolor(YELLOW);
   cprintf ("Population size:        %d\r\n",setup.population_size);
   cprintf ("Chromosome length:      %d\r\n",setup.chrom_len);
   cprintf ("Max No. of Generations: %d\r\n",setup.max_gens);
   cprintf ("Crossover Probability:  %6.4f\r\n",
     setup.crossover_prob);
   cprintf ("Mutation Probability:   %6.4f\r\n",
     setup.mutation_prob);
   cprintf ("\r\nInitial Population Statistics: \r\n\n");
   cprintf ("Max Fitness:      %10.5f\r\n",ga.max);
   cprintf ("Avg Fitness:      %10.5f\r\n",ga.avg);
   cprintf ("Min Fitness:      %10.5f\r\n",ga.min);
   cprintf ("Sum of Fitness:   %10.5f\r\n",ga.sum_fitness);
   cprintf ("\r\n       **********************************    \r\n\n");
   cprintf ("Press key to continue (esc to cancel): ");
   if (getch() == ESC_KEY) *cancel = TRUE;
   else
       *cancel = FALSE;
   return;
   }  /* end proc */

short initialize_genetic_alg (tga_setup_rec *setup,tga_rec *ga,
   float (*the_obj_funct)(ui_vector v)) {
   short opt_flag,cancel;
   int i;
   initialize_ga_rec (ga);
   ga->populations[ga->g_old] = (ppopulation)malloc (sizeof(population));
   if (!(ga->populations)[ga->g_old]) return 0;
   ga->populations[ga->g_new] = (ppopulation)malloc (sizeof(population));
   if (!(ga->populations)[ga->g_new]) return 0;
   if (!allocate_population (*setup,ga->populations[ga->g_old])) return 0;
   if (!allocate_population (*setup,ga->populations[ga->g_new])) return 0;
   ga->opt_individual.chrom =
		(short *)malloc(setup->chrom_len*sizeof(short));
   if (!(ga->opt_individual.chrom)) return 0;
   ga->allocated = 1;
   initialize_population (*setup,ga,ga->populations[ga->g_old],
       &(ga->opt_individual),the_obj_funct);
   cprintf ("\r\n\nMemory Available: %lu\r\n",(unsigned long)coreleft());
   cprintf ("\r\nDoing statistics for initial population...");
   statistics (*setup,ga->populations[ga->g_old],ga,&opt_flag);
   initial_display (*setup,*ga,&cancel);
   if (cancel) return 0;
   return 1;
   }  /* end proc */

void free_ga_pointers (tga_setup_rec setup,tga_rec *ga) {
   if (ga->allocated) {
      free_population (setup,ga->populations[0]);
      free_population (setup,ga->populations[1]);
      free_ui_vector (ga->opt_vector,0);
      }
   ga->allocated = 0;
   cprintf ("\r\nfree_ga_pointers. heapcheck = %d",heapcheck());
   getch();
   return;
   }  /*  end proc  */

void initialize_ga_setup (int pop_size,int bit_len,int vect_len,
  tga_setup_rec *the_rec) {
  int i;
  the_rec->population_size = pop_size;
  the_rec->bit_len = bit_len;
  the_rec->vect_len = vect_len;
  the_rec->chrom_len = bit_len*vect_len;
  the_rec->max_gens = 200;
  the_rec->pick_fittest = TRUE;
  the_rec->crossover_prob = 0.6;
  the_rec->mutation_prob = 1.0/pop_size;
  the_rec->tolerance = 0.01;
  return;
  }  /* end proc */

#endif
