/*  File HOPRUN.C  Run a Hopfield network  */
/*                 Connecting weights are read in from a file. */
/*                 Iterations performed 'asynchronously' */

#ifndef HOPRUN_C
#define HOPRUN_C

#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <math.h>

#include "ut.h"
#include "hop.h"

void run_hopfield_network (hop_setup_rec *the_setup,
   pattern_display_data *disp_data);

/*  Implementation */

short forward_iteration (int *iters,pattern v,pattern *v_new,int **weights,
	  int no_of_nodes) {
   /* Returns 1 if iterations converge (v_new = v), 0 otherwise */
   int i,j,sum,conv_test_sum;
   short conv = FALSE;
   float threshold = (NODE_HIGH + NODE_LOW) * 0.5;
   (*iters)++;
   *v_new = v;
   for (i = 1;i<=no_of_nodes;i++)  {
       sum = 0;
       for (j = 1;j<=no_of_nodes;j++)
	   sum +=  weights[i][j]*v_new->nodes[j-1];
       if (sum >= threshold) v_new->nodes [i-1] = NODE_HIGH;
       else v_new->nodes [i-1] = NODE_LOW;
       } /* i */
    /* test for convergence:  */
    conv_test_sum = 0;
    for (i = 0;i<no_of_nodes;i++)
	conv_test_sum += fabs(v.nodes[i] - v_new->nodes[i]);
    if (conv_test_sum == 0) conv = TRUE;
    return conv;
    } /* end proc */

void run_hopfield_network (hop_setup_rec *the_setup,
   pattern_display_data *disp_data) {
   pattern_array the_patterns;
   pattern mu,mu_new;
   int iters;
   char message[35],str_pattern_no[10],resp,resp2;
   short converged = 0;
   int pattern_no;
   int **the_weights;
   textbackground (BLUE);
   textcolor (YELLOW);
   clrscr();
   cprintf ("\r\nProgram Hoprun: Run a Hopfield neural network.");
   cprintf ("\r\nConnecting weights are read in from the file: %s.",
	 the_setup->files.in_wts_file_name);
   cprintf ("\r\nThe exemplar patterns used to create these "
	  "connecting weights");
   cprintf ("\r\nare from the file %s.",the_setup->files.data_file_name);
   cprintf ("\r\nYou can use an exemplar pattern (or a slightly modified pattern)");
   cprintf ("\r\nas a starting input pattern.  Starting patterns can also be ");
   cprintf ("\r\nbuilt up completely from scratch.");
   cprintf ("\r\nPress any key to continue (esc to cancel): ");
   if (getch() == ESC_KEY) return;

   if (!read_xmp_patterns (the_setup,disp_data,
      the_setup->files.data_file_name,the_patterns)) return;
   the_weights = allocate_2d_int_matrix (1,disp_data->no_of_nodes,
      1,disp_data->no_of_nodes);
   if (!read_hop_weights (the_setup->files.in_wts_file_name,the_weights,
      disp_data->no_of_nodes)) return;

   resp = 'y';
   while ((resp == 'y') || (resp == 'Y')) {
       clrscr();
       gotoxy (1,5);
       cprintf ("There are %d patterns in the exemplar file.",
	   disp_data->no_of_patterns);
       gotoxy (1,7);
       cprintf ("Enter the number of the exemplar pattern (1 to %d)"
	    " to be used as input: ",disp_data->no_of_patterns);
       gets (str_pattern_no);
       /* Scan for correct numeric input.  Default to 1 if incorrect */
       if (sscanf (str_pattern_no,"%d",&pattern_no) != 1) pattern_no = 1;
       if (pattern_no < 1) pattern_no = 1;
       if (pattern_no > disp_data->no_of_patterns)
	   pattern_no = disp_data->no_of_patterns;
       pattern_no--;
       mu = the_patterns[pattern_no];
       edit_pattern (disp_data,&mu,pattern_no);
       clrscr();
       gotoxy (1,5);
       cprintf ("Press any key (esc to quit) to start Hopfield iterations: ");
       if (getch() == ESC_KEY) return;
       clrscr();
       iters = 0;
       resp2 = 'y';
       while ((resp2 == 'y') || (resp2 == 'Y')) {
	   converged = forward_iteration (&iters,mu,&mu_new,the_weights,
	      disp_data->no_of_nodes);
	   clrscr();
	   display_pattern (disp_data,&mu_new,0);
	   mu = mu_new;
	   gotoxy (15,22);
	   strcpy (message,"");
	   if (converged)
	       strcpy (message,".  No change since last iteration.");
	   cprintf ("Iteration %d%s",iters,message);
	   gotoxy (15,23);
	   cprintf ("Another iteration? y/n: ");
	   resp2 = get_answer ('y','Y','n','N');
	   if (resp2 == ESC_KEY) return;
	} /* while resp2 */
	gotoxy (15,24);
	cprintf ("Press any key to clear: ");
	getch();
	clrscr();
	gotoxy (25,5);
	cprintf ("y: Process another input pattern.");
	gotoxy (25,6);
	cprintf ("e: End session.");
	gotoxy (25,8);
	cprintf ("Select option: ");
	resp = get_answer ('y','Y','e','E');
   }  /* while resp */
   textmode(LASTMODE);
   free_2d_int_matrix (the_weights,1,1);
   return;
   }  /* end proc */

#endif
