/*  File HOPTRAIN.C  Train a Hopfield network.   */
/*       Connecting weights are computed and output to disk  */

#ifndef HOPTRAIN_C
#define HOPTRAIN_C

#include <stdio.h>
#include <conio.h>
#include <string.h>

#include "hop.h"

void create_new_xmp_data_file (hop_setup_rec *the_setup,
    pattern_display_data *disp_data);
void edit_xmp_data_file (hop_setup_rec *the_setup,
   pattern_display_data *disp_data);
void train_hopfield_network (hop_setup_rec *the_setup,
   pattern_display_data *disp_data);
void print_inner_products (hop_setup_rec *the_setup,
   pattern_display_data *disp_data);

/*  Implementation  */

void create_new_xmp_data_file (hop_setup_rec *the_setup,
    pattern_display_data *disp_data) {
   pattern_array the_patterns;
   textbackground (BLUE);
   textcolor (YELLOW);
   clrscr ();
   cprintf ("\rCreate a New Exemplar Data File");
   cprintf ("\r\n\nExemplar data will be stored in file: %s",
      the_setup->files.new_data_file_name);
   cprintf ("\r\nNumber of patterns: %d.",the_setup->no_of_patterns);
   cprintf ("\r\nPattern size is %d rows by %d columns.",
      the_setup->no_of_rows,the_setup->no_of_cols);
   cprintf ("\r\n\nHit any key to continue (esc to cancel): ");
   if (getch() == ESC_KEY) return;
   disp_data->no_of_rows = the_setup->no_of_rows;
   disp_data->no_of_cols = the_setup->no_of_cols;
   disp_data->no_of_patterns = the_setup->no_of_patterns;
   disp_data->no_of_nodes = the_setup->no_of_rows * the_setup->no_of_cols;
   clrscr();
   create_new_patterns (disp_data,the_patterns);
   write_patterns (disp_data,the_setup->files.new_data_file_name,
       the_patterns);
   clrscr();
   gotoxy (1,5);
   cprintf ("\r\nNew exemplar patterns are stored in file %s.",
	     the_setup->files.new_data_file_name);
   cprintf ("\r\nPress key to clear: ");
   getch();
   textmode (LASTMODE);
   }  /* end proc */

void edit_xmp_data_file (hop_setup_rec *the_setup,
   pattern_display_data *disp_data) {
   pattern_array the_patterns;
   textbackground (BLUE);
   textcolor (YELLOW);
   clrscr();
   cprintf ("\r\nEdit Exemplar Data File");
   cprintf ("\r\n\nInput exemplar data file: %s",
	    the_setup->files.data_file_name);
   cprintf ("\r\nEdited exemplar data file: %s\r\n",
	    the_setup->files.new_data_file_name);
   if (!read_xmp_patterns (the_setup,disp_data,
      the_setup->files.data_file_name,the_patterns)) return;
   cprintf ("\r\nHit any key to continue (esc to cancel): ");
   if (getch() == ESC_KEY) return;
   clrscr();
   edit_pattern_file (disp_data,the_patterns);
   write_patterns (disp_data,the_setup->files.new_data_file_name,the_patterns);
   clrscr();
   gotoxy (1,5);
   cprintf ("\r\nEdited exemplar patterns are stored in file %s.",
	     the_setup->files.new_data_file_name);
   cprintf ("\r\nPress key to clear: ");
   getch();
   textmode (LASTMODE);
   }  /* end proc */

void train_hopfield_network (hop_setup_rec *the_setup,
   pattern_display_data *disp_data) {
   pattern_array the_patterns;
   int **the_weights;
   strcpy (the_setup->files.out_wts_file_name,
	   the_setup->files.in_wts_file_name);
   textbackground (BLUE);
   textcolor (YELLOW);
   clrscr();
   cprintf ("\r\nTrain a Hopfield neural network.");
   cprintf ("\r\nInput Exemplar File Name: %s",
	     the_setup->files.data_file_name);
   cprintf ("\r\nOutput Weights File Name: %s",
	     the_setup->files.out_wts_file_name);
   cprintf ("\r\nHit any key to continue (esc to cancel): ");
   if (getch() == ESC_KEY) return;
   if (!read_xmp_patterns (the_setup,disp_data,
	     the_setup->files.data_file_name,the_patterns)) return;
   cprintf ("\r\nHit any key to continue (esc to cancel): ");
   if (getch() == ESC_KEY) return;
   clrscr();
   the_weights = allocate_2d_int_matrix (1,disp_data->no_of_nodes,
	1,disp_data->no_of_nodes);
   compute_hop_weights (disp_data,the_patterns,the_weights);
   write_hop_weights (the_setup->files.out_wts_file_name,
       the_weights,disp_data->no_of_nodes);
   free_2d_int_matrix (the_weights,1,1);
   clrscr();
   gotoxy (1,5);
   cprintf ("Weights are stored in file %s.",
       the_setup->files.out_wts_file_name);
   cprintf ("\r\nExemplar patterns are stored in file %s.",
	     the_setup->files.data_file_name);
   cprintf ("\r\nPress key to clear: ");
   getch();
   textmode (LASTMODE);
   } /* end proc */

void print_inner_products (hop_setup_rec *the_setup,
   pattern_display_data *disp_data) {
   int i,j,k,t;
   pattern_array the_patterns;
   clrscr();
   textbackground (BLUE);
   textcolor (YELLOW);
   cprintf ("\r\nPrint inner products of exemplar patterns.\r\n");
   if (strlen(the_setup->files.data_file_name) == 0) {
      cprintf ("\r\nExemplar file name missing.");
      cprintf ("\r\nUse setup menu to enter file name.");
      cprintf ("\r\nPress key to continue: ");
      getch();
      return;
      }
   cprintf ("\r\nExemplars stored in file: %s",
       the_setup->files.data_file_name);
   cprintf ("\r\nPress key to continue (esc to cancel): ");
   if (getch() == ESC_KEY) return;
   if (!read_xmp_patterns (the_setup,disp_data,
      the_setup->files.data_file_name,the_patterns)) return;
   fprintf (stdprn,"Inner Products of Exemplar Patterns\n\n");
   fprintf (stdprn,"Exemplar Patterns stored in file %s.\n\n",
      the_setup->files.data_file_name);
   for (i = 0;i<the_setup->no_of_patterns;i++)
       for (k = 0;k<the_setup->no_of_patterns;k++) {
	   t = 0;
	   for (j = 0;j<disp_data->no_of_nodes;j++)
	       t += the_patterns[i].nodes[j]*the_patterns[k].nodes[j];
	   fprintf (stdprn,"Class %d and Class %d Inner Product = %d\n",
	       i,k,t);
	   }  /* i,k */
   }  /* end proc */
#endif