/* File PATTERNS.C Functions to read in, display, edit and store patterns */

#ifndef PATTERNS_C
#define PATTERNS_C

#include <stdio.h>
#include <alloc.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <io.h>
#include <conio.h>
#include <math.h>

#include "ut.h"
#include "pattern.h"

pattern_display_data  g_pattern_data;

short read_patterns (pattern_display_data *disp_data,path_str file_name,
      pattern_array the_patterns);
short get_pattern (pattern_display_data *disp_data,
      pattern_array the_patterns,int *pattern_no);
void display_pattern (pattern_display_data *disp_data,pattern *v,
     int pattern_no);
void edit_pattern (pattern_display_data *disp_data,pattern *v,
     int pattern_no);
void create_pattern (pattern_display_data *disp_data,pattern *v,
     int pattern_no);
void edit_pattern_file (pattern_display_data *disp_data,
     pattern_array the_patterns);
void write_patterns(pattern_display_data *disp_data,path_str file_name,
     pattern_array the_patterns);
void create_new_patterns (pattern_display_data *disp_data,
     pattern_array the_patterns);

/*  Implementation  */

short read_patterns (pattern_display_data *disp_data,path_str file_name,
    pattern_array the_patterns) {
    /* Returns 1 on successful read, 0 otherwise  */
    int	k;
    pattern the_pattern;
    int pattern_file;    /*  file of type pattern */
    int line;
    if ((pattern_file = open(file_name,O_RDONLY|O_BINARY)) == -1) {
       line = wherey() + 2;
       if (line > 24) line = 24;
       gotoxy (1,line);
       cprintf ("Problem opening file %s.",file_name);
       gotoxy (1,line + 1);
       cputs ("Hit key to continue: ");
       getch();
       return 0;
       }
    read(pattern_file,&the_pattern,MAX_NO_OF_PATTERN_NODES *
	   sizeof(int));
    disp_data->no_of_rows = the_pattern.nodes[0];
    disp_data->no_of_cols = the_pattern.nodes[1];
    disp_data->no_of_nodes = disp_data->no_of_rows *
	   disp_data->no_of_cols;
    k = 0;
    while (! eof(pattern_file)) {
	  read (pattern_file,&(the_patterns[k]),sizeof(the_pattern));
	  k++;
	  }  /* end while */
    disp_data->no_of_patterns = k;
    close (pattern_file);
    return 1;
    }	/* end function */

short get_pattern (pattern_display_data *disp_data,
   pattern_array the_patterns,int *pattern_no) {
   /*  Returns 1 if successful, 0 otherwise  */
   char  str_item[10];
   int data_item;
   do {
      gotoxy (1,5);
      cputs ("                                     ");
      gotoxy (1,5);
      cprintf ("Enter pattern number (1 to %d): ",disp_data->no_of_patterns);
      gets (str_item);
      if (str_item[0] == ESC_KEY) return 0;
      /* Scan for correct numeric input. */
      } while ((sscanf (str_item,"%d",&data_item) != 1) ||
	    ((data_item < 1) || (data_item > disp_data->no_of_patterns)));
    *pattern_no = data_item - 1;
     display_pattern (disp_data,&(the_patterns[*pattern_no]),*pattern_no);
     return 1;
     }  /* end function */

void move_to_pos (pattern_display_data *disp_data,int *j,int *i) {
   int  col,row;
   if (*j > disp_data->no_of_cols) {
      if (*i < disp_data->no_of_rows) {
	 (*i)++;
	 *j = 1;
	 }
      else *j = disp_data->no_of_cols;
      }
   if (*j < 1) {
      if (*i > 1) {
	 (*i)--;
	 *j = disp_data->no_of_cols;
	 }
       else *j = 1;
       }
   if (*i > disp_data->no_of_rows)
      *i = disp_data->no_of_rows;
   if (*i < 1) *i = 1;

   col = START_COL + 2 * (*j) - 2;
   row = START_ROW + (*i) - 1;
   gotoxy (col,row);
   return;
   } /* end proc */

void update_display_pos (pattern_display_data *disp_data,
   int value,int *col,int *row,int cursor_direction) {
   char disp_val;
   if (value == NODE_HIGH) disp_val = NODE_HIGH_DISP;
   else	 disp_val = NODE_LOW_DISP;
   move_to_pos (disp_data,col,row);
   cprintf ("%c%c",disp_val,disp_val);
   /* move cursor position */
   *col += cursor_direction;
   move_to_pos (disp_data,col,row);
   return;
   } /* end proc */

void display_pattern (pattern_display_data *disp_data,
   pattern *v,int pattern_no) {
   int	i,j,incount,row,col;
   clrscr();
   gotoxy (1,1);
   cputs ("Input Pattern: ");
   if (pattern_no > 0)
      cprintf ("Pattern %d of %d",pattern_no+1,disp_data->no_of_patterns);
   gotoxy (1,3);
   cprintf ("   %2d: %c%c    %2d: %c%c",NODE_LOW,NODE_LOW_DISP,
       NODE_LOW_DISP,NODE_HIGH,NODE_HIGH_DISP,NODE_HIGH_DISP);
   incount = 0;
   for (i = 0;i<disp_data->no_of_rows;i++) {
       for (j = 0;j< disp_data->no_of_cols;j++) {
	   col = j + 1;
	   row = i + 1;
	   update_display_pos (disp_data,v->nodes[incount],&col,&row,1);
	   incount++;
	   }  /* j */
	}  /* i */
   for (i = 1;i <= disp_data->no_of_rows;i++) {
       row = START_ROW + i - 1;
       col = START_COL - 4;
       gotoxy (col,row);
       cprintf ("%2d",i);
       } /* i */
   for (j = 1;j <= disp_data->no_of_cols;j++) {
       row = START_ROW - 2;
       col = START_COL + 2*j - 2;
       gotoxy (col,row);
       cprintf ("%2d",j);
       } /* j */
   return;
   }  /* display_pattern */

void display_pattern_and_message (pattern_display_data *disp_data,
   pattern *v,int pattern_no,int col,int row) {
   display_pattern (disp_data,v,pattern_no);
   gotoxy (col,row);
   cputs ("Arrow keys: Move around pattern.  ");
   gotoxy (col,row + 1);
   cputs ("Enter:      Change pattern element.");
   gotoxy (col,row + 2);
   cputs ("n:          Create noisy pattern.");
   gotoxy (col,row + 3);
   cputs ("q:          Quit editing this pattern.");
   return;
   }  /* end proc */

void add_noise_to_pattern (pattern_display_data *disp_data,
   pattern *v,int pattern_no) {
   const int noise_pos_x = 30,noise_pos_y = 3;
   float rnoise;
   char str_noise [10];
   int no_of_changed_nodes;
   int i,j;
   int *list_ptr;
   int x_pos,y_pos,error;
   gotoxy (noise_pos_x,noise_pos_y);
   cputs ("Enter noise percentage (0 <= noise <= 100): ");
   x_pos = wherex();
   y_pos = wherey();
   error = 0;
   do {
      gets (str_noise);
      if (sscanf (str_noise,"%f",&rnoise) == 1) error = 0;
      else error = 1;
      if (error != 0) {
	 gotoxy (x_pos,y_pos);
	 cputs ("      ");
	 gotoxy (x_pos,y_pos);
	 }
   } while (error != 0);
   if (rnoise > 0.0) {
      if (rnoise > 100.0) rnoise = 100.0;
      rnoise /= 100.0;
      no_of_changed_nodes = (int)(rnoise * disp_data->no_of_nodes);
      gotoxy (noise_pos_x,noise_pos_y);
      list_ptr = (int *)malloc (disp_data->no_of_nodes*sizeof(int));
      shuffl (disp_data->no_of_nodes,list_ptr);
      for (i = 0;i<no_of_changed_nodes;i++) {
	  if (v->nodes[list_ptr[i]] == NODE_LOW)
	      v->nodes[list_ptr[i]] = NODE_HIGH;
	  else v->nodes[list_ptr[i]] = NODE_LOW;
	  }  /* i */
      free (list_ptr);
      }
   display_pattern_and_message (disp_data,v,pattern_no,MSG_COL,MSG_ROW);
   i = 1;
   j = 1;
   move_to_pos (disp_data,&j,&i);
   return;
   }  /* end proc */

void  edit_pattern (pattern_display_data *disp_data,pattern *v,
   int pattern_no) {
   int i,j,incount,value;
   char ch;
   short quit = FALSE;
   display_pattern_and_message (disp_data,v,pattern_no,MSG_COL,MSG_ROW);
   i = 1;
   j = 1;
   move_to_pos (disp_data,&j,&i);
   while (!quit) {
      ch = getch();
      switch (ch) {
      case ENTER_KEY:
	 /* Switch pattern element value */
	 incount = (i - 1)*disp_data->no_of_cols + j;
	 if (v->nodes[incount - 1] == NODE_LOW) value = 1;
	 else value = NODE_LOW;
	 v->nodes[incount- 1] = value;
	 update_display_pos (disp_data,value,&j,&i,1);
	 break;
      case BACK_SPACE:
	 /* Back up one space and put a NODE_LOW in new position */
	 j = j - 1;
	 update_display_pos (disp_data,NODE_LOW,&j,&i,0);
	 incount = (i - 1)*disp_data->no_of_cols + j;
	 v->nodes[incount - 1] = NODE_LOW;
	 break;
      case SPACE_BAR:
	 /* Treat like right arrow */
	 j = j + 1;
	 move_to_pos(disp_data,&j,&i);
	 break;
      case 'n':
	 add_noise_to_pattern (disp_data,v,pattern_no);
	 break;
      case 'q':
	 quit = TRUE;
	 break;
      case ESC_KEY:
	 quit = TRUE;
	 break;
      case (char) 0:
	 /* Extended key: get next key code */
	 ch = getch();
	 switch (ch) {
	 case LEFT_ARROW:  j--; break;
	 case RIGHT_ARROW: j++; break;
	 case UP_ARROW:    i--; break;
	 case DOWN_ARROW:  i++; break;
	 case HOME:        j = 1; break;
	 case END_KEY:     j = disp_data->no_of_cols; break;
	 case PAGE_UP:     i = 1; break;
	 case PAGE_DOWN:   i = disp_data->no_of_rows; break;
	 case DELETE_KEY:
	     /* Put NODE_LOW in this position */
	     update_display_pos (disp_data,NODE_LOW,&j,&i,0);
	     incount = (i - 1)*disp_data->no_of_cols + j;
	     v->nodes[incount - 1] = NODE_LOW;
	     break;
	  }  /* end switch */
	  move_to_pos (disp_data,&j,&i);
	  break;  /* (char) 0 */
      } /* end switch */
      } /* end while */
    return;
    } /* end proc */

void edit_pattern_file (pattern_display_data *disp_data,
   pattern_array the_patterns) {
   char resp,ch;
   int k,i;
   short esc_flag,input_ok;
   resp = ' ';
   while (resp != 'q') {
      do {
      clrscr();
      _setcursortype(_SOLIDCURSOR);
      gotoxy (18,3);
      cputs ("a: Append a new pattern to this file.");
      gotoxy (18,4);
      cputs ("c: Change an existing pattern.");
      gotoxy (18,5);
      cputs ("d: Delete a pattern from this file.");
      gotoxy (18,6);
      cputs ("q: Quit editting of pattern file.");
      gotoxy (1,8);
      cputs ("Enter response: ");
      resp = getch();
      } while ((resp != 'a') && (resp != 'c') && (resp != 'd') &&
	    (resp != 'q') && (resp != ESC_KEY));
   if (resp == ESC_KEY) return;
   switch (resp) {
   case 'a':  /* Append new pattern */
      disp_data->no_of_patterns++;
      create_pattern (disp_data,&(the_patterns[disp_data->no_of_patterns-1]),
	  disp_data->no_of_patterns-1);
      break;  /* case 'a' */
   case 'c':  /* Change existing pattern */
      clrscr();
      gotoxy (1,3);
      cprintf ("Change a pattern: ");
      if (get_pattern (disp_data,the_patterns,&k))
	 edit_pattern (disp_data,&(the_patterns[k]),k);
      break;
   case 'd':  /* Delete a pattern */
      clrscr();
      gotoxy (1,3);
      cputs ("Delete a pattern: ");
      if (get_pattern (disp_data,the_patterns,&k)) {
	 gotoxy (1,22);
	 cputs ("Enter 'd' to delete, or esc to cancel: ");
	 do {
	   ch = getch();
	   } while ((ch != 'd') && (ch != ESC_KEY));
	 if (ch == 'd') {
	    for (i = k + 1;i<disp_data->no_of_patterns;i++)
		the_patterns[i-1] = the_patterns[i];
	    disp_data->no_of_patterns--;
	    }  /* if */
	 }  /* if */
      break;  /* case 'd' */
    }  /* switch */

  } /* while */
  clrscr();
  _setcursortype(_NORMALCURSOR);
  return;
  } /* proc */

void write_patterns(pattern_display_data *disp_data,
   path_str file_name,pattern_array the_patterns) {
   int k;
   pattern  the_pattern;
   int  pattern_file;  /* file of pattern */
   pattern_file = open (file_name,O_CREAT | O_WRONLY | O_TRUNC | O_BINARY,
      S_IREAD | S_IWRITE);
    the_pattern.nodes[0] = disp_data->no_of_rows;
    the_pattern.nodes[1] = disp_data->no_of_cols;
    write (pattern_file,&the_pattern,MAX_NO_OF_PATTERN_NODES *
	   sizeof(int));
    for (k = 0;k<disp_data->no_of_patterns;k++)
	write (pattern_file,&(the_patterns[k]),MAX_NO_OF_PATTERN_NODES *
	   sizeof(int));
   close (pattern_file);
   return;
   }  /* end proc */

void create_pattern (pattern_display_data *disp_data,
   pattern *v, int pattern_no) {
   int i;
   for (i = 0;i< disp_data->no_of_nodes;i++)
       v->nodes[i] = NODE_LOW;
   edit_pattern (disp_data,v,pattern_no);
   return;
   }  /* end proc */

void create_new_patterns (pattern_display_data *disp_data,
   pattern_array the_patterns) {
   int k;
   for (k = 0;k < disp_data->no_of_patterns;k++) {
       clrscr();
       gotoxy (1,1);
       cprintf ("Enter data for pattern %2d of %d",
		k,disp_data->no_of_patterns);
       create_pattern (disp_data,&(the_patterns[k]),k);
       } /* k */
    return;
    }  /* end proc */

#endif